#include "twi.h"
#include <zephyr/kernel.h>

/** @brief Structure containing TWIS driver instance. */
static nrfx_twis_t twis_inst = NRFX_TWIS_INSTANCE(TWIS_INST_IDX);

/** @brief Structure containing TWIM driver instance. */
static nrfx_twim_t twim_inst = NRFX_TWIM_INSTANCE(TWIM_INST_IDX);

/** @brief TWIM transmit buffer initialized with @ref MSG_TO_SEND. */
static uint8_t m_tx_buffer_master[] = MSG_TO_SEND;

/** @brief TWIS recieve buffer for a message from TWIM. */
static uint8_t m_rx_buffer_slave[sizeof(MSG_TO_SEND)];

/** @brief TWIM recieve buffer for a message from TWIS. */
static uint8_t m_rx_buffer_master[sizeof(MSG_TO_SEND)];

static uint8_t opr_config_mode[2] = {0x3d, 0x0c};
static uint8_t opr_oprmode[2] = {0x1A, 0x1B}; 
static uint8_t ret_master[2] = {0,0};


static bool m_master_done = false;

/**
 * @brief Function for handling TWIM driver events.
 *
 * @param[in] p_event   Event information structure.
 * @param[in] p_context General purpose parameter set during initialization of the TWIM.
 *                      This parameter can be used to pass additional information to the
 *                      handler function. In this example @p p_context is used to pass address
 *                      of TWI transfer descriptor structure.
 */
static void twim_handler(nrfx_twim_evt_t const * p_event, void * p_context)
{
    nrfx_err_t status;
    (void)status;

    nrfx_twim_xfer_desc_t * twim_desc = p_context;
    //static bool m_master_done = false;

    if (p_event->type == NRFX_TWIM_EVT_DONE)
    {
        // NRFX_LOG_INFO("--> Master event: done - transfer completed");
        // twim_desc->type = NRFX_TWIM_XFER_RX;
        // twim_desc->p_primary_buf = m_rx_buffer_master;

        // if (!m_master_done)
        // {
        //     status = nrfx_twim_xfer(&twim_inst, twim_desc, 0);
        //     NRFX_ASSERT(status == NRFX_SUCCESS);
        //     m_master_done = true;
        // }
        // else
        // {
        //     // NRFX_LOG_INFO("Content of master RX buffer: %s", m_rx_buffer_master);
        //     NRFX_LOG_INFO("Content of master RX buffer: 0x%x, 0x%x, 0x%x, 0x%x", m_rx_buffer_master[0],m_rx_buffer_master[1],m_rx_buffer_master[2],m_rx_buffer_master[3]);
        // }
        m_master_done = true;
        if(twim_desc->type == NRFX_TWIM_XFER_RX)
        {
            printf("Content of master RX buffer: 0x%x, 0x%x\n", m_rx_buffer_master[0],m_rx_buffer_master[1]);
        }
    }
    else
    {
        printf("--> MASTER handler, event: %d.\n", p_event->type);
    }
}

/**
 * @brief Function for handling TWIS driver events.
 *
 * @param[in] p_event Event information structure.
 */
static void twis_handler(nrfx_twis_evt_t const * p_event)
{
    nrfx_err_t status;
    (void)status;

    switch (p_event->type)
    {
        case NRFX_TWIS_EVT_WRITE_DONE:
            printf("--> Slave event: write done.\n");
            printf("--> 0x%x, 0x%x, 0x%x, 0x%x\n",m_rx_buffer_slave[0],m_rx_buffer_slave[1],m_rx_buffer_slave[2],m_rx_buffer_slave[3]);
            break;

        case NRFX_TWIS_EVT_READ_DONE:
            printf("--> Slave event: read done.\n");
            break;

        case NRFX_TWIS_EVT_WRITE_REQ:
            status = nrfx_twis_rx_prepare(&twis_inst, m_rx_buffer_slave, sizeof(m_rx_buffer_slave));
            NRFX_ASSERT(status == NRFX_SUCCESS);
            printf("--> Slave event: write request\n");
            break;

        case NRFX_TWIS_EVT_READ_REQ:
            // status = nrfx_twis_tx_prepare(&twis_inst, m_rx_buffer_slave, sizeof(m_rx_buffer_slave));
            ret_master[0]++;
            status = nrfx_twis_tx_prepare(&twis_inst, ret_master, sizeof(ret_master));
            NRFX_ASSERT(status == NRFX_SUCCESS);
            printf("--> Slave event: read request\n");

            if(ret_master[0] == 0xFF)
            {
                ret_master[0] = 0;
                ret_master[1]++;
                if(ret_master[1] == 0xFF)
                    ret_master[1] = 0;
            }

            break;

        default:
            printf("--> SLAVE event: %d.\n", p_event->type);
    }
}

/**
 * @brief Function for application main entry.
 *
 * @return Nothing.
 */
int twi_init(void)
{
    nrfx_err_t status;
    (void)status;

    printf("Starting nrfx_twim_twis non-blocking example.\n");

    // nrfx_twim_xfer_desc_t twim_xfer_desc = NRFX_TWIM_XFER_DESC_TX(SLAVE_ADDR,
    //                                                               m_tx_buffer_master,
    //                                                               sizeof(m_tx_buffer_master));

    nrfx_twim_xfer_desc_t twim_xfer_desc = NRFX_TWIM_XFER_DESC_TX(SLAVE_ADDR,opr_config_mode,2);

    nrfx_twim_config_t twim_config = NRFX_TWIM_DEFAULT_CONFIG(MASTER_SCL_PIN, MASTER_SDA_PIN);
    status = nrfx_twim_init(&twim_inst, &twim_config, twim_handler, &twim_xfer_desc);
    NRFX_ASSERT(status == NRFX_SUCCESS);

    nrfx_twis_config_t twis_config = NRFX_TWIS_DEFAULT_CONFIG(SLAVE_SCL_PIN,
                                                              SLAVE_SDA_PIN,
                                                              SLAVE_ADDR);
    status = nrfx_twis_init(&twis_inst, &twis_config, twis_handler);
    NRFX_ASSERT(status == NRFX_SUCCESS);

#if defined(__ZEPHYR__)
    IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_TWIM_INST_GET(TWIM_INST_IDX)), IRQ_PRIO_LOWEST,
                       NRFX_TWIM_INST_HANDLER_GET(TWIM_INST_IDX), 0);

    IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_TWIS_INST_GET(TWIS_INST_IDX)), IRQ_PRIO_LOWEST,
                       NRFX_TWIS_INST_HANDLER_GET(TWIS_INST_IDX), 0);
#endif

    // NRFX_LOG_INFO("Message to send: %s", m_tx_buffer_master);
    // NRFX_LOG_INFO("Content of master RX buffer: %s", m_rx_buffer_master);
    // NRFX_EXAMPLE_LOG_PROCESS();

    printf("opr_config_mode[0]: 0x%x,opr_config_mode[1]: 0x%x\n", opr_config_mode[0],opr_config_mode[1]);

    nrfx_twis_enable(&twis_inst);
    nrfx_twim_enable(&twim_inst);

    status = nrfx_twim_xfer(&twim_inst, &twim_xfer_desc, 0);
    NRFX_ASSERT(status == NRFX_SUCCESS);

    k_sleep(K_MSEC(50));

    printf("opr_oprmode[0]: 0x%x,opr_oprmode[1]: 0x%x\n", opr_oprmode[0],opr_oprmode[1]);

    twim_xfer_desc.p_primary_buf    = opr_oprmode;
    twim_xfer_desc.primary_length   = 1;

    status = nrfx_twim_xfer(&twim_inst, &twim_xfer_desc, 0);
    NRFX_ASSERT(status == NRFX_SUCCESS);

    k_sleep(K_MSEC(5));

    while (1)
    {
        if(m_master_done != false)
        {
            m_master_done = false;
            twim_xfer_desc.type             = NRFX_TWIM_XFER_RX;
            twim_xfer_desc.primary_length   = 2;
            twim_xfer_desc.p_primary_buf    = m_rx_buffer_master;
            nrfx_twim_xfer(&twim_inst, &twim_xfer_desc, 0);
            k_sleep(K_SECONDS(1));
        }
    }
}

/** @} */
